/*************************************************************************************************
 *
 * Modeling and animation (TNM079) 2007
 * Code base for lab assignments. Copyright:
 *   Gunnar Johansson (gunnar.johansson@itn.liu.se)
 *   Ken Museth (ken.museth@itn.liu.se)
 *   Michael Bang Nielsen (bang@daimi.au.dk)
 *   Ola Nilsson (ola.nilsson@itn.liu.se)
 *   Andreas Sderstrm (andreas.soderstrom@itn.liu.se)
 *
 *************************************************************************************************/
#include "HalfEdgeMesh.h"
#include "ColorMap.h"
#ifdef __APPLE__
#include "GLUT/glut.h"
#else
#include "GL/glut.h"
#endif

//contains classes for defining the numerical 
//data type limits on each computer platform
#include <limits>


const unsigned int HalfEdgeMesh::BORDER = std::numeric_limits<unsigned int>::max();
const unsigned int HalfEdgeMesh::UNINITIALIZED = std::numeric_limits<unsigned int>::max()-1;

HalfEdgeMesh::HalfEdgeMesh()
{
  setShadingFlag(SMOOTH_SHADING);
}

HalfEdgeMesh::~HalfEdgeMesh()
{
}



//-----------------------------------------------------------------------------
bool HalfEdgeMesh::addTriangle(const Vector3<float> &v1, const Vector3<float> &v2, const Vector3<float> & v3)
{

  // Add the vertices of the face/triangle
  unsigned int ind1, ind2, ind3;
  addVertex(v1, ind1);
  addVertex(v2, ind2);
  addVertex(v3, ind3);

  // Add all half-edge pairs
  
  //the edge indexes that will be used 
  //to connect  the edge pair
  unsigned int edgeInd1,edgeInd2,edgeInd3,edgeInd4,edgeInd5,edgeInd6;
  bool border1,border2,border3;

  //the first edge pair is betwen  the vertex indices v1 and v2
  border1 = addHalfEdgePair(ind1,ind2,edgeInd1,edgeInd2);


  //the second edge pair is between  the vertex indices v2 and v3
  border2 = addHalfEdgePair(ind2,ind3,edgeInd3,edgeInd4);


  //the third edge pair is between the vertex indices v3 and v1
  border3 = addHalfEdgePair(ind3,ind1,edgeInd5,edgeInd6);

  //set  the border for the halfedge
  if(border1)
  {
		mEdges.at(edgeInd2).face = BORDER;
  }

  if(border2)
  {
  		mEdges.at(edgeInd4).face = BORDER;
  }

  if(border3)
  {
		mEdges.at(edgeInd6).face = BORDER;
  }


  // Connect inner ring
  // that connect  the next 
  // and the inner edges
  
  //the edges adjacent to v2 are now connected
  mEdges.at(edgeInd1).next = edgeInd3;
  mEdges.at(edgeInd3).prev = edgeInd1;

  //the edges adjacent to v3 are now connected
  mEdges.at(edgeInd3).next = edgeInd5;
  mEdges.at(edgeInd5).prev = edgeInd3;


  //the edges adjacent to v1 are noe connected
  mEdges.at(edgeInd5).next = edgeInd1;
  mEdges.at(edgeInd1).prev = edgeInd5;


  //Finally create the face and add
  //that to  the mesh
  Face face;

  //the face list only need to store 
  //the index to one of the half-edges
  //that borders it - doesnt really matter
  //which one
  face.edge = edgeInd1;

  mFaces.push_back(face);


  //All inner loop half-edges share the same
  //left face(previosly added) 
  //connect them to  the face

  mEdges.at(edgeInd1).face = mFaces.size() - 1;
  mEdges.at(edgeInd3).face = mFaces.size() - 1;
  mEdges.at(edgeInd5).face = mFaces.size() - 1;

  /**Optionally, track  the outer boundary half-edges 
	* to represent non-closed surfaces
	*/

  return true;


}



//-----------------------------------------------------------------------------
bool HalfEdgeMesh::addVertex(const Vector3<float> & v, unsigned int &indx)
{
  std::map<Vector3<float>,unsigned int>::iterator it = mUniqueVerts.find(v);

  if (it != mUniqueVerts.end())
  {
    indx = (*it).second;
    return false;
  }

  mUniqueVerts[v] = indx = mVerts.size(); // op. [ ] constructs a new entry in map
  Vertex vert;
  vert.vec = v;
  mVerts.push_back(vert);
    
  return true;
}


bool HalfEdgeMesh::addHalfEdgePair(unsigned int v1, unsigned int v2, unsigned int &indx1, unsigned int &indx2)
{
  // Search for the HalfEdge pair among existing edges
  std::vector<HalfEdge>::reverse_iterator iter = mEdges.rbegin();
  std::vector<HalfEdge>::reverse_iterator iend = mEdges.rend();

  while (iter != iend) 
  {
    if ((*iter).vert == v1 && mEdges[(*iter).pair].vert == v2) 
	 {
      indx1 = iter.base() - mEdges.begin() - 1;
      indx2 = (*iter).pair;
      return false;
    }
    iter++;
  }

  // If not found, add both half-edges
  indx1 = mEdges.size();
  indx2 = indx1+1;

  // Create edges and set pair index
  HalfEdge edge1, edge2;
  edge1.pair = indx2;
  edge2.pair = indx1;

  // Connect the edges to the verts
  edge1.vert = v1;
  edge2.vert = v2;

  // Connect the verts to the edges
  mVerts[v1].edge = indx1;
  mVerts[v2].edge = indx2;

  // Store the edges
  mEdges.push_back(edge1);
  mEdges.push_back(edge2);

  return true;
}

/**Updates the boundary of the mesh to contain the 
 * HalfEdgeMesh::HalfEdge pointed to by indx 
 * \param [in] indx the index to  the half-edge,
 * unsigned int
 */
void HalfEdgeMesh::mergeBoundaryEdge(unsigned int indx)
{

   //initialize a halfEdge with 
	//index that is passed as the parameter
	HalfEdge &he = mEdges.at(indx);
 	
	he.next = mEdges.at(mEdges.at(he.pair).prev).pair;

   
	while(mEdges.at(he.next).face != BORDER)
	{
		he.next = mEdges.at(mEdges.at(he.next).prev).pair;
	}

	mEdges.at(he.next).prev = indx;



	he.prev = mEdges.at(mEdges.at(he.pair).next).pair;

	while(mEdges.at(he.prev).face != BORDER)
	{
		he.prev = mEdges.at(mEdges.at(he.prev).next).pair;
	} 
	
	mEdges.at(he.prev).next = indx;

}


void HalfEdgeMesh::mergeAllBoundaries()
{
      for(unsigned int i = 0;i < mEdges.size(); i++)
		{
			if(mEdges.at(i).face == BORDER)
			{
				mergeBoundaryEdge(i);
			}
		}
}


void HalfEdgeMesh::validate()
{
  std::vector<HalfEdge>::iterator iterEdge = mEdges.begin();
  std::vector<HalfEdge>::iterator iterEdgeEnd = mEdges.end();
  while (iterEdge != iterEdgeEnd) 
  {
    if ((*iterEdge).face == UNINITIALIZED ||
        (*iterEdge).next == UNINITIALIZED ||
        (*iterEdge).pair == UNINITIALIZED ||
        (*iterEdge).prev == UNINITIALIZED ||
        (*iterEdge).vert == UNINITIALIZED)
        std::cerr << "HalfEdge " << iterEdge - mEdges.begin() << " not properly initialized" << std::endl;

    iterEdge++;
  }
  std::cerr << "Done with edge check (checked " << mEdges.size() << " edges)" << std::endl;

  std::vector<Face>::iterator iterFace = mFaces.begin();
  std::vector<Face>::iterator iterFaceEnd = mFaces.end();
  while (iterFace != iterFaceEnd) 
  {
    if ((*iterFace).edge == UNINITIALIZED)
        std::cerr << "Face " << iterFace - mFaces.begin() << " not properly initialized" << std::endl;

    iterFace++;
  }
  std::cerr << "Done with face check (checked " << mFaces.size() << " faces)" << std::endl;

  std::vector<Vertex>::iterator iterVertex = mVerts.begin();
  std::vector<Vertex>::iterator iterVertexEnd = mVerts.end();
  while (iterVertex != iterVertexEnd) 
  {
    if ((*iterVertex).edge == UNINITIALIZED)
        std::cerr << "Vertex " << iterVertex - mVerts.begin() << " not properly initialized" << std::endl;

    iterVertex++;
  }
  std::cerr << "Done with vertex check (checked " << mVerts.size() << " vertices)" << std::endl;

  std::cerr << "Looping through triangle neighborhood of each vertex... ";
  iterVertex = mVerts.begin();
  iterVertexEnd = mVerts.end();
  std::vector<unsigned int> foundTriangles;
  int emptyCount = 0;

  while (iterVertex != iterVertexEnd) 
  {
    if (!findNeighbourTriangles(iterVertex - mVerts.begin(), foundTriangles))
      emptyCount++;

    iterVertex++;
  }
  std::cerr << std::endl << "Done: " << emptyCount << " isolated vertices found" << std::endl;


}


//-----------------------------------------------------------------------------
bool HalfEdgeMesh::findNeighbourTriangles(const unsigned int vertexIndex, std::vector<unsigned int>& foundTriangles) const
{
  foundTriangles.clear();

  //Find othe triangles that include that vertex
  //which is simply the 1-ring neighborhood
  /*
  std::vector<Face>::const_iterator iter = mFaces.begin();
  std::vector<Face>::const_iterator iend = mFaces.end();

  while(iter != iend)
  {
		if ( mEdges.at((*iter).edge).vert == vertexIndex ||
		     mEdges.at(mEdges.at((*iter).edge).next).vert == vertexIndex ||
			  mEdges.at(mEdges.at((*iter).edge).prev).vert == vertexIndex 
		   )
		{
			  foundTriangles.push_back(iter - mFaces.begin());
		}

		iter++;
  }
  */

  //retrieve the edge index with vertex index
  unsigned int edgeInd  = mVerts.at(vertexIndex).edge;

  do
  {
		if(mEdges.at(edgeInd).face != BORDER)
		{
			foundTriangles.push_back(mEdges.at(edgeInd).face);
		}

		edgeInd = mEdges.at(mEdges.at(edgeInd).prev).pair;

  }while(edgeInd != mVerts.at(vertexIndex).edge);

  return !(foundTriangles.size() == 0);			

}


//calculate the area of the mesh
float HalfEdgeMesh::area() const
{

  unsigned int numTriangles = mFaces.size();
  float area = 0.0;
  
  Vector3<float> p0,p1,p2,n,v1,v2;

  for (unsigned int i = 0; i < numTriangles; i++)
  {

	 p0 = mVerts.at(mEdges.at(mFaces.at(i).edge).vert).vec; 
	
	 p1 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).next).vert).vec;

	 p2 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).prev).vert).vec;


	 //calculate the face normal
	 v1 = p1 - p0;
	 v2 = p2 - p0;

	 n = cross(v1,v2);

	 area  += (0.5 *  n.length());

  }	

  return area; 
}

float HalfEdgeMesh::volume() const
{
//   std::cout << "Inside  the volume calculation " << std::endl; 

	Vector3<float> p0,p1,p2,n,addedVertices,area_normal,v1,v2;

	float volume = 0.0, area = 0.0;

	unsigned int numTriangles = mFaces.size();

	for(unsigned int i = 0;i < numTriangles; i++)
	{
	    //retrieve all the vertices of the triangles
		 p0 = mVerts.at(mEdges.at(mFaces.at(i).edge).vert).vec; 
	
	    p1 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).next).vert).vec;

	    p2 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).prev).vert).vec;	

		 //add the retrieved vertices and divide by 3
       addedVertices = (p0 + p1 + p2)/3;


		 //retrieve the face normal of that triangle
		 n = mFaces.at(i).normal;

		 //calculate the area of the i-th triangle
	    v1 = p1 - p0;
	    v2 = p2 - p0;

	    area_normal = cross(v1,v2);	 

		 area = 0.5 * area_normal.length();

		/*
		 std::cout << "The area of the sphere is: " << area << std::endl;
		 std::cout << "added vertices :" << addedVertices << std::endl;
		 std::cout << "face normal;" << n << std::endl;
		 */
		 volume += addedVertices * n * area;
	}

   return volume/3;
}

int HalfEdgeMesh::genus() const
{
  //get  the total number of vertices
  unsigned int numUniqueVertices = mVerts.size();

  //get the total number of faces
  unsigned int numFaces = mFaces.size();

  //get  the total number of edges
  unsigned int numEdges = mEdges.size()/2;

//  std::cout << "Genus: " << (2 - numFaces + numEdges - numUniqueVertices)/2 << std::endl;

  //the shell has to be at least 1
  unsigned int shell = 1,genus = 0;

  //the number of loops is the same as the number of faces
  //as there is exactly one loop for each face
  unsigned int loop = mFaces.size();

  //declare a STL queue data structure
  //that will store the vertex index
  std::queue<unsigned int> vertexIndexQueue;

  //declare a STL set data structure
  //that will store the vertex index
  std::set<unsigned int> vertexIndexSet;


  //versIterator represents pair containing const_iterator
  //and bool
  std::set<unsigned int>::const_iterator  vertexIndexIterator;

  //the following STL vector keeps track of the 
  //location adjacent faces to a vertex
  std::vector<unsigned int> neighbourTriangles;


  //looping through  the vertitex array
  for (unsigned int v = 0; v < numUniqueVertices; v++)
  {

		//insert  the vertex index
		//in the vertex index queue
		vertexIndexQueue.push(v);

	   while(!vertexIndexQueue.empty())
		{
			//insert the vertex index in the set 
			//that is popped off from the queue
			//as we know that STL set takes only 
			//unique value duplicates will be discarded
			vertexIndexSet.insert(vertexIndexQueue.front());

			vertexIndexQueue.pop();

			//the following function takes the vertex index as a 
	   	//parameter and find the location index of the adjacent
	   	//triangles to that vertex
	   	findNeighbourTriangles(v,neighbourTriangles);		

			unsigned int numOfCandidates = neighbourTriangles.size();
		
			for(unsigned int i = 0;i < numOfCandidates; i++)
			{
				//retrieve the neighbouring vertex index 
				unsigned int neighbourVertexIndex = mEdges.at(mFaces.at(neighbourTriangles.at(i)).edge).vert;
			   //look for the item in the set
				vertexIndexIterator = vertexIndexSet.find(neighbourVertexIndex);

				//if the item - which  is the 
				//vertex index is found
				//means DUPLICATES
				if(vertexIndexIterator != vertexIndexSet.end())
				{
						//std::cout << "Already in the set" << std::endl;
				}
				else
				{
					vertexIndexQueue.push(neighbourVertexIndex);
				}
			}
		}


  }
  //now we compare the size of the VertexIndexSet 
  //between the size of mVerts
  if(vertexIndexSet.size() == mVerts.size())
  {
			shell = 1;

			genus =static_cast<unsigned int> ( (2 - numFaces + numEdges - numUniqueVertices)/2);
  }
  else
  {
			std::cout << "Something is wrong that has to be figured out" << std::endl;
  }
  //here you calculate the genus
  //genus = static_cast<unsigned int>(shell - 0.5*(numUniqueVertices - numEdges + numFaces));
  /*
  std::cout << "The size of the unique vertex map is: " << mUniqueVerts.size() << std::endl;
  std::cout << "The size of the vertex vector is: " << mVerts.size() << std::endl; 
  std::cout << "The size of the set is: " << vertexIndexSet.size() << std::endl;
  std::cout << "The size of the queue is: " << vertexIndexQueue.size() << std::  endl; 
  */
  return genus;
}

float HalfEdgeMesh::curvature(const unsigned int vertexIndex, const Vector3<float>& n)
{
  //printf("Curvature calculation not implemented for half-edge mesh!\n");
  
  

  return 0;
}

void HalfEdgeMesh::calculateFaceNormals()
{

  unsigned int numTriangles = mFaces.size();

  //set the shading flag
  setShadingFlag(FLAT_SHADING);

  Vector3<float> n;

  for (unsigned int i = 0; i < numTriangles; i++)
  {

    // implement the normal calculation here

	 Vector3<float>& p0 = mVerts.at(mEdges.at(mFaces.at(i).edge).vert).vec; 
	
	 Vector3<float>& p1 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).next).vert).vec;

	 Vector3<float>& p2 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).prev).vert).vec;


	 //calculate the face normal
	 Vector3<float> v1 = p1 - p0;
	 Vector3<float> v2 = p2 - p0;

	 n = cross(v1,v2);
	 n.normalize();


    mFaces[i].normal = n; // assign normal to face
  }	
}

void HalfEdgeMesh::calculateVertexNormals()
{

  //set  the shading flag
  setShadingFlag(SMOOTH_SHADING);

  //the following STL vector keeps track of the 
  //location adjacent faces to a vertex
  std::vector<unsigned int> neighbourTriangles;

  unsigned int numVerts = mVerts.size();
  
  for (unsigned int v = 0; v < numVerts; v++)
  {
    //the following function takes the vertex index as a 
	 //parameter and find the location index of the adjacent
	 //triangles to that vertex
	 findNeighbourTriangles(v,neighbourTriangles);

	 //std::cout << "the nighbour are: " << neighbourTriangles.size() << std::endl;

	 //Approaximated vertex normal
	 Vector3<float> n(0,0,0);


	 const unsigned int numCandidates = neighbourTriangles.size();

	 for(unsigned int i = 0;i < numCandidates; i++) 
	 {
		n += mFaces.at(neighbourTriangles.at(i)).normal;			
	 }

	 n.normalize();

    // assign normal to vertex
	 mVerts.at(v).v_normal = n;
  }	
  
}


//-----------------------------------------------------------------------------
void HalfEdgeMesh::draw() 
{
  ColorMap colorMap;
	
  glMatrixMode(GL_MODELVIEW);
	
  // Apply transform	
  glPushMatrix(); // Push modelview matrix onto stack
	
  // Convert transform-matrix to format matching GL matrix format
  GLfloat glMatrix[16];
  mTransform.toGLMatrix(glMatrix); 
	
  // Load transform into modelview matrix 	
  glLoadMatrixf(glMatrix);
	
  // Draw geometry	
  glBegin(GL_TRIANGLES);
  const int numTriangles = mFaces.size();
  for (int i = 0; i < numTriangles; i++)
  {

    HalfEdge* edge = &mEdges[mFaces[i].edge];
	
    Vector3<float>& p0 = mVerts[edge->vert].vec;
    edge = &mEdges[edge->next];

    Vector3<float>& p1 = mVerts[edge->vert].vec;
    edge = &mEdges[edge->next];

    Vector3<float>& p2 = mVerts[edge->vert].vec;
		
    if (getShadingFlag() == FLAT_SHADING)
	 {
      Vector3<float>& n = mFaces[i].normal;
      Vector3<float> color = colorMap.map(n, -1, 1);
						
      glColor3f(color[0],color[1], color[2]); 			
      glNormal3f(n[0], n[1], n[2]);
      glVertex3f(p0[0], p0[1], p0[2]);
      glVertex3f(p1[0], p1[1], p1[2]);
      glVertex3f(p2[0], p2[1], p2[2]);
    }
    else if (getShadingFlag() == SMOOTH_SHADING)
	 {

      // First you need to calculate the vertex normals in calculateVertexNormals
      // Then you need to get correct vertex normals here
      //std::cerr << "Vertex normal shading in HalfEdgeMesh::draw() not implemented. ";
      Vector3<float> n0, n1, n2;
	   
		n0 = mVerts.at(mEdges.at(mFaces.at(i).edge).vert).v_normal;

		n1 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).next).vert).v_normal;

		n2 = mVerts.at(mEdges.at(mEdges.at(mFaces.at(i).edge).prev).vert).v_normal;

      // Color mapping, maps the normal components to R,G,B. From [-1, 1] to [0,1] respectively
      Vector3<float> color0 = colorMap.map(n0, -1, 1);
      Vector3<float> color1 = colorMap.map(n1, -1, 1);
      Vector3<float> color2 = colorMap.map(n2, -1, 1);
      
      glColor3f(color0[0],color0[1], color0[2]); 			
      glNormal3f(n0[0], n0[1], n0[2]);
      glVertex3f(p0[0], p0[1], p0[2]);
						
      glColor3f(color1[0],color1[1], color1[2]); 				 		
      glNormal3f(n1[0], n1[1], n1[2]);
      glVertex3f(p1[0], p1[1], p1[2]);
						
      glColor3f(color2[0],color2[1], color2[2]); 				 		
      glNormal3f(n2[0], n2[1], n2[2]);
      glVertex3f(p2[0], p2[1], p2[2]);					
    }
    else
	 {
      // No normals
      glVertex3f(p0[0], p0[1], p0[2]);
      glVertex3f(p1[0], p1[1], p1[2]);
      glVertex3f(p2[0], p2[1], p2[2]);			
    }		
		
  }  
  glEnd();
	
  // Restore modelview matrix
  glPopMatrix();
}
